# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from datetime import timedelta
from hashlib import sha1
from logging import getLogger
from pathlib import Path
from random import shuffle
from shutil import rmtree
from tempfile import gettempdir
from time import gmtime, sleep, strftime, time
from typing import Any, Dict, Iterator, List, Optional
from urllib.parse import urlsplit

from ffpuppet import BrowserTimeoutError, Debugger, FFPuppet, LaunchError, Reason

from .reporter import FuzzManagerReporter

LOG = getLogger(__name__)
TMP_PATH = Path(gettempdir()) / "site-scout"
TMP_PATH.mkdir(exist_ok=True, parents=True)


class Status:
    """Track and report status to a file."""

    __slots__ = ("_dst", "_next", "_rate_limit", "_start")

    def __init__(self, dst: Path, rate_limit: int = 60):
        assert dst
        assert rate_limit >= 0
        self._dst = dst
        self._next: float = 0
        self._rate_limit = rate_limit
        self._start = time()

    def report(
        self,
        active: int,
        jobs: int,
        completed: int,
        target: int,
        results: int,
        force: bool = False,
    ) -> None:
        """Write status report to the filesystem.

        Args:
            active: Number of active browsers.
            jobs: Maximum number of active browsers.
            completed: URLs opened.
            target: Total URLs to be opened.
            results: Number of results found.
            force: Ignore rate limit and report.

        Returns:
            None
        """
        now = time()
        if not force and self._next > now:
            return
        comp_pct = (completed / target) * 100 if target else 0.0
        with self._dst.open("w") as lfp:
            lfp.write(f" Active/Limit : {active}/{jobs}\n")
            lfp.write(f"Current/Total : {completed}/{target} ({comp_pct:0.1f}%)\n")
            lfp.write(f"      Results : {results}\n")
            lfp.write(f"      Runtime : {timedelta(seconds=int(now - self._start))}\n")
            lfp.write(f"    Timestamp : {strftime('%Y/%m/%d %X %z', gmtime())}\n")
        self._next = now + self._rate_limit


class URL:
    """URL components."""

    __slots__ = ("domain", "scheme", "subdomain", "path", "_uid")

    def __init__(
        self,
        domain: str,
        subdomain: Optional[str] = None,
        path: str = "/",
        scheme: str = "http",
    ):
        assert domain
        assert path.startswith("/")
        assert subdomain is None or subdomain
        self.domain = domain.lower()
        self.scheme = scheme.lower()
        assert self.scheme in ("http", "https")
        self.subdomain = subdomain.lower() if subdomain else None
        self.path = path
        self._uid: Optional[str] = None

    def __str__(self) -> str:
        if self.subdomain is None:
            return f"{self.scheme}://{self.domain}{self.path}"
        return f"{self.scheme}://{self.subdomain}.{self.domain}{self.path}"

    @property
    def uid(self) -> str:
        """Unique ID.

        Args:
            None

        Returns:
            Unique ID.
        """
        # this does NOT need to be cryptographically secure
        # it needs to be filesystem safe and *somewhat* unique
        if not self._uid:
            self._uid = sha1(str(self).encode(errors="replace")).hexdigest()
        return self._uid


class Visit:
    """Visit contains details about the site and browser."""

    __slots__ = ("idle_timestamp", "puppet", "url", "timestamp")

    def __init__(self, puppet: FFPuppet, url: URL, timestamp: Optional[float] = None):
        self.idle_timestamp: Optional[float] = None
        self.puppet = puppet
        self.url = url
        self.timestamp = time() if timestamp is None else timestamp
        assert self.timestamp >= 0


# pylint: disable=too-many-instance-attributes
class SiteScout:
    """SiteScout can visit a collection of URLs and report process failures."""

    __slots__ = (
        "_active",
        "_binary",
        "_cert_files",
        "_complete",
        "_debugger",
        "_display",
        "_extension",
        "_fuzzmanager",
        "_launch_timeout",
        "_log_limit",
        "_memory_limit",
        "_prefs",
        "_profile",
        "_urls",
    )

    def __init__(
        self,
        binary: Path,
        profile: Optional[Path] = None,
        prefs_js: Optional[Path] = None,
        debugger: Debugger = Debugger.NONE,
        display: Optional[str] = None,
        launch_timeout: int = 180,
        log_limit: int = 0,
        memory_limit: int = 0,
        extension: Optional[List[Path]] = None,
        cert_files: Optional[List[Path]] = None,
        fuzzmanager: bool = False,
    ):
        self._active: List[Visit] = []
        self._complete: List[Visit] = []
        self._urls: List[URL] = []
        # browser related
        self._binary = binary
        self._cert_files = cert_files
        self._debugger = debugger
        self._extension = extension
        self._display = display
        self._launch_timeout = launch_timeout
        self._log_limit = log_limit
        self._memory_limit = memory_limit
        self._prefs = prefs_js
        self._profile = profile
        # reporter
        self._fuzzmanager: Optional[FuzzManagerReporter] = (
            FuzzManagerReporter(binary, working_path=TMP_PATH) if fuzzmanager else None
        )

    def __enter__(self) -> "SiteScout":
        return self

    def __exit__(self, *exc: Any) -> None:
        self.close()

    def close(self) -> None:
        """Close and cleanup.

        Args:
            None

        Returns:
            None
        """
        LOG.debug("closing %d active visit(s)...", len(self._active))
        for visit in self._active:
            visit.puppet.clean_up()

    def _launch(self, url: URL, launch_attempts: int = 3) -> None:
        """Launch a new browser instance and visit provided URL.

        Args:
            url: URL to visit.
            launch_attempts: Attempts to launch the browser before raising.

        Returns:
            None
        """
        assert launch_attempts > 0

        env_mod: Dict[str, Optional[str]] = {"MOZ_CRASHREPORTER_SHUTDOWN": "1"}

        for attempt in range(1, launch_attempts + 1):
            ffp = FFPuppet(
                debugger=self._debugger,
                headless=self._display,
                use_profile=self._profile,
                working_path=str(TMP_PATH),
            )
            try:
                ffp.launch(
                    self._binary,
                    env_mod=env_mod,
                    location=str(url),
                    launch_timeout=self._launch_timeout,
                    log_limit=self._log_limit,
                    memory_limit=self._memory_limit,
                    prefs_js=self._prefs,
                    extension=self._extension,
                    cert_files=self._cert_files,
                )
            except LaunchError as exc:
                ffp.clean_up()
                LOG.warning(
                    "Browser launch %s (attempt %d/%d)",
                    "timeout" if isinstance(exc, BrowserTimeoutError) else "failure",
                    attempt,
                    launch_attempts,
                )
                if attempt == launch_attempts:
                    raise
                # launch attempt limit not met... retry
                sleep(1)
                continue
            break

        self._active.append(Visit(ffp, url))

    def load_dict(self, data: Dict[str, Dict[str, List[str]]]) -> None:
        """Load URLs from a dict add to list of URLs to visit.

        Args:
            data: Dictionary containing URLs.

        Returns:
            None
        """
        total_domains = 0
        total_subdomains = 0
        total_urls = 0
        while data:
            domain, subdomains = data.popitem()
            total_domains += 1
            for subdomain, paths in subdomains.items():
                total_subdomains += 1
                for path in paths:
                    # "*" indicates no subdomain
                    url = URL(
                        domain,
                        subdomain=subdomain if subdomain != "*" else None,
                        path=path,
                    )
                    LOG.debug("-> '%s'", url)
                    # this might get slow with large lists
                    if url.uid not in (x.uid for x in self._urls):
                        self._urls.append(url)
                    total_urls += 1
        LOG.debug(
            "%d domain(s), %d subdomain(s), %d URL(s) processed",
            total_domains,
            total_subdomains,
            total_urls,
        )

    def load_str(self, url: str) -> None:
        """Parse, sanitize and add a URL to list of URLs to visit.

        Args:
            url: Location to visit.

        Returns:
            None
        """
        assert url
        if "://" in url:
            scheme = url.split("://", maxsplit=1)[0]
            if scheme.lower() not in ("http", "https"):
                raise ValueError(f"Unsupported scheme in URL: {scheme}")
        else:
            url = f"http://{url}"
        parsed = urlsplit(url, allow_fragments=False)
        path = parsed.path if parsed.path else "/"
        if parsed.query:
            path = f"{path}?{parsed.query}"
        # this currently does not separate domain and subdomain
        formatted = URL(parsed.netloc, path=path, scheme=parsed.scheme)
        # this might get slow with large lists
        if formatted.uid not in (x.uid for x in self._urls):
            self._urls.append(formatted)

    def _process_active(
        self,
        time_limit: float,
        idle_usage: int = 10,
        idle_wait: float = 10.0,
        min_visit: int = 20,
    ) -> None:
        """Check active browser processes and move complete visits from the active
        to the complete queue.

        Args:
            time_limit: Maximum time to wait for a site to load.
            idle_usage: Maximum CPU usage to be considered idle.
            idle_wait: Amount of time in seconds process must be idle.
            min_visit: Delay in seconds before idle checks begin.

        Returns:
            None
        """
        assert time_limit > 0
        assert min_visit > 0
        assert idle_usage >= 0
        assert idle_wait >= 0

        complete: List[int] = []
        for index, visit in enumerate(self._active):
            # check if work is complete
            visit_runtime = time() - visit.timestamp
            if not visit.puppet.is_healthy():
                visit.puppet.close()
                complete.append(index)
            elif visit_runtime >= time_limit:
                LOG.debug("visit timeout (%s)", visit.url.uid[:8])
                visit.puppet.close()
                complete.append(index)
            # check all browser processes are below idle limit
            elif idle_usage and visit_runtime >= min_visit:
                if all(x[1] < idle_usage for x in visit.puppet.cpu_usage()):
                    now = time()
                    if visit.idle_timestamp is None:
                        LOG.debug("set idle (%s)", visit.url.uid[:8])
                        visit.idle_timestamp = now
                    if now - visit.idle_timestamp >= idle_wait:
                        LOG.debug("visit idle (%s)", visit.url.uid[:8])
                        visit.puppet.close()
                        complete.append(index)
                elif visit.idle_timestamp is not None:
                    LOG.debug("reset idle (%s)", visit.url.uid[:8])
                    visit.idle_timestamp = None

        if complete:
            LOG.debug("found %d complete", len(complete))
            for index in sorted(complete, reverse=True):
                self._complete.append(self._active.pop(index))
            LOG.debug("%d active, %d complete", len(self._active), len(self._complete))

    def _report(self, log_path: Path, logs_only: bool = False) -> Iterator[Path]:
        """Save available results to a provided location and cleanup browser data.

        Args:
            log_path: Directory to save results in.
            logs_only: Skip extra debugger results.

        Yields:
            Directory containing a results.
        """
        while self._complete:
            LOG.debug("%d pending visit(s) to check", len(self._complete))
            visit = self._complete.pop()
            if visit.puppet.reason in (Reason.ALERT, Reason.WORKER):
                dst = log_path / strftime(f"%Y%m%d-%H%M%S-result-{visit.url.uid[:8]}")
                visit.puppet.save_logs(dst, logs_only=logs_only)
                if self._prefs:
                    (dst / "prefs.js").write_text(self._prefs.read_text())
                (dst / "url.txt").write_text(str(visit.url))
                yield dst
            visit.puppet.clean_up()

    def schedule_urls(self, url_limit: int = 0, randomize: bool = True) -> None:
        """Prepare URL list. Randomize and limit size as needed.

        Args:
            url_limit: Limit total URLs when value is greater than zero.
            shuffle_urls: Randomly order URLs visits.

        Returns:
            None
        """
        assert url_limit >= 0
        if randomize:
            shuffle(self._urls)
        if url_limit and len(self._urls) > url_limit:
            LOG.info("Enforcing URL limit (%d -> %d)", len(self._urls), url_limit)
            self._urls = self._urls[:url_limit]

    # pylint: disable=too-many-locals
    def run(
        self,
        log_path: Path,
        time_limit: int,
        check_delay: float = 1.0,
        domain_rate_limit: int = 20,
        instance_limit: int = 1,
        status_report: Optional[Path] = None,
    ) -> None:
        """Iterate over and visit each URL. Each visit is performed in a new browser
        instance using a clean profile.

        Args:
            log_path: Location to write results.
            time_limit: Maximum time in seconds of a site visit.
            check_delay: Time in seconds between checking for results.
            domain_rate_limit: Minimum time is seconds between visiting the same domain.
            instance_limit: Maximum number of browser sessions to run at once.
            status_report: File to populate with status report data.

        Returns:
            None
        """
        assert check_delay >= 0
        assert domain_rate_limit >= 0
        assert instance_limit > 0

        last_visit: Dict[str, float] = {}
        next_url: Optional[URL] = None
        status = Status(status_report) if status_report else None
        total_results = 0
        total_urls = len(self._urls)
        total_visits = 0
        while self._urls or self._active:
            assert not next_url

            # perform status report
            if status:
                status.report(
                    len(self._active),
                    instance_limit,
                    total_visits,
                    total_urls,
                    total_results,
                )

            # select url to visit
            if self._urls and (len(self._active) < instance_limit):
                next_url = self._urls.pop()
                # avoid hammering specific domains by using rate limit
                if time() - last_visit.get(next_url.domain, 0) < domain_rate_limit:
                    LOG.debug("domain rate limit hit (%s)", next_url.domain)
                    self._urls.insert(0, next_url)
                    next_url = None

            # launch browser and visit url
            if next_url:
                url_str = str(next_url)
                total_visits += 1
                LOG.info(
                    "[%02d/%02d] %s %r",
                    total_visits,
                    total_urls,
                    next_url.uid[:8],
                    f"{url_str[:80]}..." if len(url_str) > 80 else url_str,
                )
                self._launch(next_url)
                last_visit[next_url.domain] = time()
                next_url = None
                assert self._active

            # check for complete processes
            self._process_active(time_limit)

            # report/save results
            for result in self._report(log_path):
                total_results += 1
                LOG.info("Result found! (%d)", total_results)
                if self._fuzzmanager:
                    fm_id, short_sig = self._fuzzmanager.submit(result)
                    LOG.info(
                        "FuzzManager (%d): %s",
                        fm_id,
                        f"{short_sig[:50]}..." if len(short_sig) > 50 else short_sig,
                    )
                    rmtree(result)
                else:
                    LOG.info("Saved as '%s'", result)

            # wait a moment for work to complete
            if self._urls or self._active:
                sleep(check_delay)

        # final status report
        if status:
            status.report(
                len(self._active),
                instance_limit,
                total_visits,
                total_urls,
                total_results,
                force=True,
            )
        LOG.info("URLs visited %d, results reported %d", total_visits, total_results)


# pylint: disable=too-many-return-statements
def verify_dict(data: Any) -> Optional[str]:
    """Verify the structure of data.

    Args:
        data: Dictionary to check.

    Return:
        An error message is returned if an problem is found.
    """
    if not isinstance(data, dict):
        return "Invalid data"
    if not data:
        return "No data found"
    # check domains
    for domain, subdomains in data.items():
        if not isinstance(domain, str) or not domain:
            return "Domain must be a string"
        if not isinstance(subdomains, dict) or not subdomains:
            return f"Invalid domain entry: '{domain}'"
        # check subdomains
        for subdomain, paths in subdomains.items():
            if not isinstance(subdomain, str) or not subdomain:
                return "Subdomain must be a string"
            if not isinstance(paths, list) or not paths:
                return f"Invalid subdomain entry: '{subdomain}' in '{domain}'"
            # check paths
            for path in paths:
                if not isinstance(path, str) or not path.startswith("/"):
                    return "Path must be a string starting with '/'"
    return None



# type: ignore
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
from pytest import mark

from .main import generate_prefs, load_input, main


@mark.parametrize(
    "args",
    [
        # only required args
        [],
        # log-level
        ["--log-level", "DEBUG"],
        # url
        ["-u", "http://mozilla.com"],
    ],
)
def test_main_01(mocker, tmp_path, args):
    """test main()"""
    ffp = mocker.patch("site_scout.site_scout.FFPuppet", autospec=True)
    ffp.return_value.is_healthy.return_value = False
    fake_bin = tmp_path / "fake_browser_bin"
    fake_bin.touch()
    default_args = [str(fake_bin)]
    # -i or -u is required
    if "-u" not in args:
        data = tmp_path / "data.yml"
        data.write_text("{'d':{'s':['/']}}")
        default_args.extend(["-i", str(data)])
    main(default_args + args)


def test_main_02(mocker, tmp_path):
    """test main() KeyboardInterrupt"""
    mocker.patch(
        "site_scout.main.SiteScout", autospec=True, side_effect=KeyboardInterrupt()
    )
    fake_bin = tmp_path / "fake_browser_bin"
    fake_bin.touch()
    main([str(fake_bin), "-u", "a.b.c"])


def test_generate_prefs_01(tmp_path):
    """test generate_prefs()"""
    assert generate_prefs(dst=tmp_path).is_file()


def test_load_input_01(tmp_path):
    """test load_input()"""
    # empty list
    assert not list(load_input([]))
    # empty directory
    assert not list(load_input([tmp_path]))

    valid = "{'d':{'s':['/']}}"
    # single input file
    in_file = tmp_path / "sites.yml"
    in_file.write_text(valid)
    results = list(load_input([in_file]))
    assert len(results) == 1
    # multiple input files
    results = list(load_input([in_file, in_file]))
    assert len(results) == 2
    # single input directory
    results = list(load_input([tmp_path]))
    assert len(results) == 1

    # invalid file data
    (tmp_path / "ignore.txt").write_text("foo")
    (tmp_path / "bad.yml").write_text("{-")
    (tmp_path / "sites.yml").write_text(valid)
    (tmp_path / "a.yml").write_text("foo")
    results = list(load_input([tmp_path]))
    assert len(results) == 1



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from logging import DEBUG, basicConfig, getLogger
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Dict, Iterator, List, Optional

from prefpicker import PrefPicker
from yaml import safe_load
from yaml.parser import ParserError
from yaml.scanner import ScannerError

from .args import parse_args
from .site_scout import SiteScout, verify_dict

LOG = getLogger(__name__)


def init_logging(level: int) -> None:
    """Initialize logging

    Arguments:
        level: logging verbosity level

    Returns:
        None
    """
    if level == DEBUG:
        date_fmt = "%m-%d %H:%M:%S"
        log_fmt = "%(asctime)s.%(msecs)03d %(levelname).1s %(name)s | %(message)s"
    else:
        date_fmt = "%H:%M:%S"
        log_fmt = "[%(asctime)s] %(message)s"
    basicConfig(format=log_fmt, datefmt=date_fmt, level=level)


def generate_prefs(dst: Optional[Path] = None, variant: str = "a11y") -> Path:
    """Generate prefs.js file via PrefPicker.

    Arguments:
        variant: Variant to use.

    Returns:
        pref.js file.
    """
    with NamedTemporaryFile(prefix="scout-prefs-", suffix=".js", dir=dst) as tmp_fp:
        prefs = Path(tmp_fp.name)
    picker = PrefPicker.load_template(
        Path(__file__).parent / "resources" / "prefpicker.yml"
    )
    picker.create_prefsjs(prefs, variant=variant)
    return prefs


def load_input(src: List[Path]) -> Iterator[Dict[str, Dict[str, List[str]]]]:
    """Load data from filesystem.

    Arguments:
        src: Files and directories to load data from.

    Yields:
        URL data.
    """
    for entry in scan_input(src):
        LOG.debug("loading '%s'", entry)
        with entry.open("r") as in_fp:
            try:
                data = safe_load(in_fp)
            except (ParserError, ScannerError):
                LOG.warning("Load failure - Invalid yml (ignored: %s)", entry)
                continue
        err_msg = verify_dict(data)
        if err_msg:
            LOG.warning("Load failure - %s (ignored: %s)", err_msg, entry)
            continue
        yield data


def scan_input(src: List[Path]) -> Iterator[Path]:
    """Scan list of inputs which can include files or directories.

    Arguments:
        src: Paths the evaluate.

    Yields:
        URL files to load.
    """
    for entry in src:
        if entry.resolve().is_dir():
            yield from entry.glob("*.yml")
        else:
            yield entry


def main(argv: Optional[List[str]] = None) -> None:
    """Main function"""
    args = parse_args(argv)
    assert any(args.input) != any(args.url)
    init_logging(args.log_level)

    tmp_prefs = False
    try:
        # generate prefs.js file if one is not provided
        if args.prefs is None:
            args.prefs = generate_prefs()
            tmp_prefs = True

        LOG.info("Starting Site Scout...")
        with SiteScout(
            args.binary,
            profile=args.profile,
            prefs_js=args.prefs,
            debugger=args.debugger,
            display="default" if args.display == "headless" else args.display,
            launch_timeout=args.launch_timeout,
            log_limit=args.log_limit,
            memory_limit=args.memory_limit,
            extension=None,
            cert_files=None,
            fuzzmanager=args.fuzzmanager,
        ) as scout:
            for data in load_input(args.input):
                scout.load_dict(data)
            for in_url in args.url:
                scout.load_str(in_url)
            # don't randomize urls passed on the command line
            scout.schedule_urls(url_limit=args.url_limit, randomize=any(args.input))
            scout.run(
                args.output_path,
                args.time_limit,
                instance_limit=args.jobs,
                status_report=args.status_report,
            )
    except KeyboardInterrupt:
        LOG.warning("Aborting...")

    finally:
        if tmp_prefs:
            args.prefs.unlink()
        LOG.info("Done.")



# type: ignore
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
# pylint: disable=missing-docstring,protected-access
from itertools import chain, count, cycle, repeat

from ffpuppet import LaunchError, Reason
from pytest import mark, raises

from .site_scout import URL, SiteScout, Status, verify_dict


@mark.parametrize(
    "domain, subdomain, path, scheme, expected",
    [
        # with a subdomain
        ("a.c", "b", "/d", "http", "http://b.a.c/d"),
        # without a subdomain
        ("a.c", None, "/d", "http", "http://a.c/d"),
        # more complex
        ("a.c:1337", "x.y", "/1/2/3.html", "https", "https://x.y.a.c:1337/1/2/3.html"),
        # normalize case (domain and subdomain only)
        ("aB.cD", "Ab", "/Ef", "HTTP", "http://ab.ab.cd/Ef"),
        # normalize case (missing subdomain)
        ("aB.cD", None, "/Ef", "HTTP", "http://ab.cd/Ef"),
    ],
)
def test_url_01(domain, subdomain, path, scheme, expected):
    """test URL()"""
    url = URL(domain, subdomain=subdomain, path=path, scheme=scheme)
    assert str(url) == expected
    assert url.uid


def test_site_scout_launch(mocker):
    """test SiteScout._launch()"""
    mocker.patch("site_scout.site_scout.FFPuppet", autospec=True)
    with SiteScout(None) as scout:
        assert not scout._active
        scout._launch("http://someurl/")
        assert scout._active
        assert scout._active[0].idle_timestamp is None
        assert scout._active[0].puppet is not None
        assert scout._active[0].timestamp > 0
        assert scout._active[0].url == "http://someurl/"


@mark.parametrize(
    "attempts, failures, success",
    [
        # single attempt failure
        (1, 1, False),
        # success on second attempt
        (2, 1, True),
    ],
)
def test_site_scout_launch_retries(mocker, attempts, failures, success):
    """test SiteScout._launch()"""
    mocker.patch("site_scout.site_scout.sleep", autospec=True)
    ffp = mocker.patch("site_scout.site_scout.FFPuppet", autospec=True)
    ffp.return_value.launch.side_effect = chain(
        repeat(LaunchError(), failures), cycle([None])
    )
    with SiteScout(None) as scout:
        assert not scout._active
        if success:
            scout._launch("http://a/", launch_attempts=attempts)
        else:
            with raises(LaunchError):
                scout._launch("http://a/", launch_attempts=attempts)
        assert ffp.return_value.clean_up.call_count == failures


@mark.parametrize(
    "urls, is_healthy, timeout, cpu_usage, idle, active",
    [
        # no urls to process
        ([], True, False, None, False, 0),
        # one active
        ([URL("foo")], True, False, None, False, 1),
        # multiple active
        ([URL("foo"), URL("bar")], True, False, None, False, 2),
        # one complete
        ([URL("foo")], False, False, None, False, 0),
        # timeout
        ([URL("foo")], True, True, None, False, 0),
        # idle
        ([URL("foo")], True, False, 0, False, 0),
        # reset idle
        ([URL("foo")], True, False, 100, True, 1),
    ],
)
def test_site_scout_process_active(
    mocker, urls, is_healthy, timeout, cpu_usage, idle, active
):
    """test SiteScout._process_active()"""
    ffpuppet = mocker.patch("site_scout.site_scout.FFPuppet", autospec=True)
    ffpuppet.return_value.is_healthy.return_value = is_healthy
    ffpuppet.return_value.cpu_usage.return_value = [(None, cpu_usage)]
    with SiteScout(None) as scout:
        assert not scout._active
        for url in urls:
            scout._launch(url)
        total_active = len(scout._active)
        # setup state
        if timeout:
            for visit in scout._active:
                visit.timestamp = 0
        if cpu_usage is not None:
            for visit in scout._active:
                visit.timestamp -= 10
                if idle:
                    visit.idle_timestamp = visit.timestamp
        # run and verify
        scout._process_active(30, idle_usage=10, idle_wait=0, min_visit=5)
        assert len(scout._active) == active
        assert active or total_active == len(scout._complete)
        for entry in scout._active:
            assert not entry.puppet.close.call_count
        for entry in scout._complete:
            assert entry.puppet.close.call_count


@mark.parametrize(
    "urls, reason, reports",
    [
        # no urls to process
        ([], None, 0),
        # interesting result
        ([URL("foo")], Reason.ALERT, 1),
        # multiple interesting results
        ([URL("foo"), URL("bar")], Reason.ALERT, 2),
        # uninteresting result
        ([URL("foo")], Reason.CLOSED, 0),
        # uninteresting result
        ([URL("foo")], Reason.EXITED, 0),
        # interesting result
        ([URL("foo")], Reason.WORKER, 1),
    ],
)
def test_site_scout_report(mocker, tmp_path, urls, reason, reports):
    """test SiteScout._report()"""

    # pylint: disable=unused-argument
    def save_logs(dst_path, logs_only=False):
        dst_path.mkdir(exist_ok=True)

    ffpuppet = mocker.patch("site_scout.site_scout.FFPuppet", autospec=True)
    ffpuppet.return_value.is_healthy.return_value = False
    ffpuppet.return_value.reason = reason
    ffpuppet.return_value.save_logs.side_effect = save_logs
    prefs = tmp_path / "prefs.js"
    prefs.touch()
    report_dst = tmp_path / "reports"
    report_dst.mkdir()
    with SiteScout(None, prefs_js=prefs) as scout:
        assert not scout._active
        for url in urls:
            scout._launch(url)
        assert scout._active or not urls
        scout._process_active(30)
        assert not scout._active
        assert sum(1 for _ in scout._report(report_dst)) == reports
    assert sum(1 for _ in report_dst.iterdir()) == reports


@mark.parametrize(
    "urls, reason, jobs, reports, use_fm, status",
    [
        # no urls to process
        ([], None, 1, 0, False, True),
        # interesting result
        ([URL("foo")], Reason.ALERT, 1, 1, False, False),
        # job > work
        ([URL("foo")], Reason.ALERT, 2, 1, False, False),
        # multiple interesting results
        ([URL("foo"), URL("bar")], Reason.ALERT, 1, 2, False, False),
        # work > jobs
        ([URL("1"), URL("2"), URL("3"), URL("4")], Reason.ALERT, 2, 4, False, False),
        # uninteresting result
        ([URL("foo")], Reason.CLOSED, 1, 0, False, False),
        # domain rate limit
        ([URL("foo"), URL("foo")], Reason.CLOSED, 1, 0, False, False),
        # timeout
        ([URL("foo")], None, 1, 0, False, False),
        # report via FuzzManager
        ([URL("foo")], Reason.ALERT, 1, 1, True, False),
        # report status
        ([URL("foo")], Reason.ALERT, 1, 1, False, True),
    ],
)
def test_site_scout_run(mocker, tmp_path, urls, reason, jobs, reports, use_fm, status):
    """test SiteScout.run()"""

    # pylint: disable=unused-argument
    def save_logs(dst_path, logs_only=False):
        dst_path.mkdir(exist_ok=True)

    mocker.patch("site_scout.site_scout.sleep", autospec=True)
    mocker.patch("site_scout.site_scout.time", autospec=True, side_effect=count())
    ffpuppet = mocker.patch("site_scout.site_scout.FFPuppet", autospec=True)
    reporter = mocker.patch("site_scout.site_scout.FuzzManagerReporter", autospec=True)
    reporter.return_value.submit.return_value = (1337, "[@ sig]")
    if reason:
        # only first pass is running
        ffpuppet.return_value.is_healthy.side_effect = chain([True], cycle([False]))
    ffpuppet.return_value.reason = reason
    ffpuppet.return_value.save_logs.side_effect = save_logs

    report_dst = tmp_path / "reports"
    report_dst.mkdir()
    status_file = tmp_path / "status.txt" if status else None
    with SiteScout(None, fuzzmanager=use_fm) as scout:
        assert not scout._active
        scout._urls = urls
        scout.run(report_dst, 10, instance_limit=jobs, status_report=status_file)
    assert sum(1 for _ in report_dst.iterdir()) == (0 if use_fm else reports)
    assert reporter.return_value.submit.call_count == (reports if use_fm else 0)


@mark.parametrize(
    "urls, input_data",
    [
        # no urls to process
        ([], {}),
        # load urls
        (["http://a.c/", "http://b.a.c/d"], {"a.c": {"*": ["/"], "b": ["/d"]}}),
        # normalizing
        (["http://b.a.c/D"], {"A.C": {"B": ["/D"]}}),
    ],
)
def test_site_scout_load_dict(urls, input_data):
    """test SiteScout.load_dict()"""
    with SiteScout(None) as scout:
        assert not scout._active
        scout.load_dict(input_data)
        for url in scout._urls:
            assert str(url) in urls
        assert len(urls) == len(scout._urls)


@mark.parametrize(
    "url, result",
    [
        # domain and tld (missing scheme)
        ("a.b", "http://a.b/"),
        # subdomain, domain and tld
        ("a.b.c", "http://a.b.c/"),
        # with scheme https
        ("https://a.b.c/", "https://a.b.c/"),
        # with scheme http
        ("http://a.b.c", "http://a.b.c/"),
        # with port
        ("a.b:1234", "http://a.b:1234/"),
        # port and path
        ("a.b:1234/c", "http://a.b:1234/c"),
        # port, path, parameters, query and fragment
        ("a.b/c;p?q=1&q2#f", "http://a.b/c;p?q=1&q2#f"),
        # normalizing domain and scheme
        ("HTTPS://A.B.C/eFg1", "https://a.b.c/eFg1"),
    ],
)
def test_site_scout_load_str_01(url, result):
    """test SiteScout.load_str()"""
    with SiteScout(None) as scout:
        scout.load_str(url)
        assert scout._urls
        assert result in str(scout._urls[0])


def test_site_scout_load_str_02():
    """test SiteScout.load_str() invalid scheme"""
    with SiteScout(None) as scout:
        with raises(ValueError):
            scout.load_str("ftp://a.b.c")


def test_site_scout_load_collision():
    """test loading an existing URL"""
    existing = "a.b.c"
    with SiteScout(None) as scout:
        scout.load_str(existing)
        assert len(scout._urls) == 1
        scout.load_str(existing)
        assert len(scout._urls) == 1
        scout.load_dict({"b.c": {"a": ["/"]}})
        assert len(scout._urls) == 1


@mark.parametrize(
    "active, jobs, completed, target, results, force",
    [
        # nothing running
        (0, 1, 0, 0, 0, False),
        # running with single site to visit
        (1, 1, 0, 1, 0, False),
        # running with single site to visit, forced report
        (1, 1, 0, 1, 0, True),
        # typical scenario
        (2, 3, 4, 10, 1, False),
    ],
)
def test_site_scout_status(
    mocker, tmp_path, active, jobs, completed, target, results, force
):
    """test Status()"""
    mocker.patch(
        "site_scout.site_scout.time", autospec=True, side_effect=count(start=1)
    )
    dst = tmp_path / "status.txt"
    status = Status(dst, rate_limit=2)
    assert status
    assert status._next == 0
    status.report(active, jobs, completed, target, results, force=force)
    assert dst.is_file()
    assert status._next > 0
    next_report = status._next
    status.report(active, jobs, completed, target, results, force=force)
    if not force:
        assert status._next == next_report
    else:
        assert status._next > next_report


@mark.parametrize(
    "size, limit, randomize",
    [
        # empty
        (0, 0, True),
        # empty with limit
        (0, 1, True),
        # no limit
        (10, 0, True),
        # enforce limit
        (10, 2, True),
    ],
)
def test_site_scout_schedule_urls(size, limit, randomize):
    """test Status.schedule_urls()"""
    with SiteScout(None) as scout:
        # prepare scout._urls
        scout._urls = list(range(size))
        scout.schedule_urls(url_limit=limit, randomize=randomize)
        if limit and size >= limit:
            assert len(scout._urls) == limit
        else:
            assert len(scout._urls) == size


@mark.parametrize(
    "data, msg",
    [
        # empty
        ({}, "No data found"),
        # not a dict
        ([], "Invalid data"),
        # valid
        ({"d": {"s": ["/"]}}, None),
        # empty domain name
        ({"": {"s": ["/"]}}, "Domain must be a string"),
        # empty domain entry
        ({"d": {}}, "Invalid domain entry: 'd'"),
        # empty subdomain name
        ({"d": {"": ["/"]}}, "Subdomain must be a string"),
        # empty subdomain entry
        ({"d": {"s": []}}, "Invalid subdomain entry: 's' in 'd'"),
        # empty path entry
        ({"d": {"s": [""]}}, "Path must be a string starting with '/'"),
    ],
)
def test_verify_dict(data, msg):
    """test verify_dict()"""
    assert verify_dict(data) == msg



# type: ignore
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
from pytest import raises

from .args import parse_args


def test_args_01(tmp_path):
    """test parse_args()"""
    dummy_file = tmp_path / "dummy_file"
    dummy_file.touch()
    assert parse_args([str(dummy_file), "-i", str(dummy_file)])


def test_args_02(capsys, tmp_path):
    """test parse_args() checks"""
    dummy_file = tmp_path / "dummy_file"
    dummy_file.touch()

    with raises(SystemExit):
        parse_args(["missing", "-i", str(dummy_file)])
    assert "binary does not exist: 'missing'" in capsys.readouterr()[-1]

    with raises(SystemExit):
        parse_args([str(dummy_file), "-i", "missing"])
    assert "-i/--input does not exist: 'missing'" in capsys.readouterr()[-1]

    with raises(SystemExit):
        parse_args([str(dummy_file), "-i", str(dummy_file), "-o", "missing"])
    assert "-o/--output-path does not exist: 'missing'" in capsys.readouterr()[-1]

    with raises(SystemExit):
        parse_args([str(dummy_file), "-i", str(dummy_file), "--prefs", "missing"])
    assert "--prefs does not exist: 'missing'" in capsys.readouterr()[-1]

    with raises(SystemExit):
        parse_args([str(dummy_file), "-i", str(dummy_file), "--jobs", "0"])
    assert "--jobs must be >= 1" in capsys.readouterr()[-1]

    with raises(SystemExit):
        parse_args([str(dummy_file), "-i", str(dummy_file), "--log-limit", "-1"])
    assert "--log-limit must be >= 0" in capsys.readouterr()[-1]

    with raises(SystemExit):
        parse_args([str(dummy_file), "-i", str(dummy_file), "--memory-limit", "-1"])
    assert "--memory-limit must be >= 0" in capsys.readouterr()[-1]

    with raises(SystemExit):
        parse_args([str(dummy_file), "-i", str(dummy_file), "--time-limit", "-1"])
    assert "--time-limit must be > 0" in capsys.readouterr()[-1]


def test_args_03(mocker, capsys, tmp_path):
    """test parse_args() rr/pernosco checks"""
    mocker.patch("site_scout.args.system", autospec=True, return_value="Linux")
    fake_which = mocker.patch("site_scout.args.which", autospec=True)
    dummy_file = tmp_path / "dummy_file"
    dummy_file.touch()

    with raises(SystemExit):
        parse_args([str(dummy_file), "-u", "foo", "--rr", "--fuzzmanager"])
    assert "rr not supported with --fuzzmanager" in capsys.readouterr()[-1]

    fake_which.return_value = None
    with raises(SystemExit):
        parse_args([str(dummy_file), "-u", "foo", "--rr"])
    assert "rr is not installed" in capsys.readouterr()[-1]

    fake_which.return_value = "rr"
    mocker.patch("site_scout.args.Path.read_bytes", autospec=True, return_value=b"99")
    with raises(SystemExit):
        parse_args([str(dummy_file), "-u", "foo", "--rr"])
    assert "perf_event_paranoid <= 1, but it is 99" in capsys.readouterr()[-1]



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from argparse import ArgumentParser, Namespace
from importlib.metadata import PackageNotFoundError, version
from logging import DEBUG, INFO
from os import getenv
from pathlib import Path
from platform import system
from shutil import which
from typing import List, Optional

from ffpuppet import Debugger

try:
    __version__ = version("site-scout")
except PackageNotFoundError:  # pragma: no cover
    # package is not installed
    __version__ = "unknown"

TIME_LIMIT_DEBUG = 300
TIME_LIMIT_DEFAULT = 120


def is_headless() -> bool:
    """Detect if running in a headless environment.

    Args:
        None

    Returns:
        True if running in a headless environment otherwise False.
    """
    return (
        system() == "Linux" and not getenv("DISPLAY") and not getenv("WAYLAND_DISPLAY")
    )


# pylint: disable=too-many-branches,too-many-statements
def parse_args(argv: Optional[List[str]] = None) -> Namespace:
    """Argument parsing"""
    parser = ArgumentParser(
        prog="site-scout", description="Visit provided URLs and report results."
    )
    parser.add_argument("binary", type=Path, help="Firefox binary to use.")

    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        "-i",
        "--input",
        default=[],
        nargs="+",
        type=Path,
        help="File(s) containing URLs to visit.",
    )
    group.add_argument("-u", "--url", default=[], nargs="+", help="URL(s) to visit.")

    display_choices = ["headless"]
    if system() == "Linux":
        display_choices.append("xvfb")
    parser.add_argument(
        "--display",
        choices=display_choices,
        default="xvfb" if is_headless() else None,
        help="Display mode.",
    )
    parser.add_argument(
        "--fuzzmanager", action="store_true", help="Report results to FuzzManager."
    )
    parser.add_argument(
        "--jobs", type=int, default=1, help="Maximum number of browsers to run at once."
    )
    parser.add_argument(
        "--launch-timeout",
        type=int,
        default=300,
        help="Time in seconds to wait for browser to launch and begin navigation"
        " (default: %(default)s).",
    )
    level_map = {"INFO": INFO, "DEBUG": DEBUG}
    parser.add_argument(
        "--log-level",
        choices=sorted(level_map),
        default="INFO",
        help="Configure console logging (default: %(default)s).",
    )
    parser.add_argument(
        "--log-limit",
        type=int,
        default=0,
        help="Browser log file size limit in MBs (default: no limit).",
    )
    parser.add_argument(
        "--memory-limit",
        type=int,
        default=0,
        help="Browser memory limit in MBs (default: no limit).",
    )
    parser.add_argument(
        "--prefs",
        type=Path,
        help="Custom prefs.js file to use (default: generated).",
    )
    parser.add_argument("--profile", type=Path, help="")
    parser.add_argument(
        "-o",
        "--output-path",
        default=Path.cwd(),
        type=Path,
        help="Location to save results (default: %(default)s).",
    )
    parser.add_argument(
        "--status-report",
        type=Path,
        help="Location to save periodic status report (default: no report).",
    )
    parser.add_argument(
        "--time-limit",
        type=int,
        help=f"Page load time limit in seconds (default: {TIME_LIMIT_DEFAULT}).",
    )
    parser.add_argument(
        "--url-limit",
        type=int,
        default=0,
        help="Maximum number of URLs to visit (default: no limit).",
    )
    parser.add_argument(
        "--version",
        "-V",
        action="version",
        version=f"%(prog)s {__version__}",
        help="Show version number.",
    )

    parser.set_defaults(debugger=Debugger.NONE)
    if system() == "Linux":
        dbg_group = parser.add_mutually_exclusive_group()
        dbg_group.add_argument(
            "--pernosco",
            action="store_const",
            const=Debugger.PERNOSCO,
            dest="debugger",
            help="Use rr. Trace intended to be used with Pernosco.",
        )
        dbg_group.add_argument(
            "--rr",
            action="store_const",
            const=Debugger.RR,
            dest="debugger",
            help="Use rr.",
        )

    args = parser.parse_args(argv)

    if args.time_limit is None:
        if args.debugger == Debugger.NONE:
            args.time_limit = TIME_LIMIT_DEFAULT
        else:
            args.time_limit = TIME_LIMIT_DEBUG
    if args.time_limit < 1:
        parser.error("--time-limit must be > 0 (recommended minimum: 30)")

    if args.debugger in (Debugger.PERNOSCO, Debugger.RR):
        if args.fuzzmanager:
            parser.error("rr not supported with --fuzzmanager")
        # rr is only supported on Linux
        if not which("rr"):
            parser.error("rr is not installed")
        settings = "/proc/sys/kernel/perf_event_paranoid"
        value = int(Path(settings).read_bytes())
        if value > 1:
            parser.error(f"rr needs {settings} <= 1, but it is {value}")

    if args.jobs < 1:
        parser.error("--jobs must be >= 1")

    args.log_level = level_map[args.log_level]

    if args.log_limit < 0:
        parser.error("--log-limit must be >= 0")
    args.log_limit *= 1_048_576

    if args.memory_limit < 0:
        parser.error("--memory-limit must be >= 0")
    args.memory_limit *= 1_048_576

    if not args.binary.is_file():
        parser.error(f"binary does not exist: '{args.binary}'")

    for in_path in args.input:
        if not in_path.exists():
            parser.error(f"-i/--input does not exist: '{in_path}'")

    if not args.output_path.is_dir():
        parser.error(f"-o/--output-path does not exist: '{args.output_path}'")

    if args.prefs and not args.prefs.is_file():
        parser.error(f"--prefs does not exist: '{args.prefs}'")

    return args
